@@ -15,7 +15,6 @@ import alphafold2_pytorch.constants as constants
 
 # structure module
 
-from en_transformer import EnTransformer
 from invariant_point_attention import IPATransformer
 
 # constants
@@ -27,13 +26,6 @@ Logits = namedtuple('Logits', ['distance', 'theta', 'phi', 'omega'])
 def exists(val):
     return val is not None
 
-def maybe(fn):
-    def inner(t, *args, **kwargs):
-        if not exists(t):
-            return None
-        return fn(t, *args, **kwargs)
-    return inner
-
 def default(val, d):
     if exists(val):
         return val
@@ -108,10 +100,12 @@ class Attention(nn.Module):
 
         self.dropout = nn.Dropout(dropout)
 
-    def forward(self, x, mask = None, attn_bias = None, **kwargs):
-        device, orig_shape, h = x.device, x.shape, self.heads
+    def forward(self, x, mask = None, attn_bias = None, context = None, context_mask = None):
+        device, orig_shape, h, has_context = x.device, x.shape, self.heads, exists(context)
 
-        q, k, v = (self.to_q(x), *self.to_kv(x).chunk(2, dim = -1))
+        context = default(context, x)
+
+        q, k, v = (self.to_q(x), *self.to_kv(context).chunk(2, dim = -1))
 
         i, j = q.shape[-2], k.shape[-2]
 
@@ -130,8 +124,9 @@ class Attention(nn.Module):
 
         if exists(mask):
             mask = default(mask, lambda: torch.ones(1, i, device = device).bool())
+            context_mask = mask if not has_context else default(context_mask, lambda: torch.ones(1, k.shape[-2], device = device).bool())
             mask_value = -torch.finfo(dots.dtype).max
-            mask = mask[:, None, :, None] * mask[:, None, None, :]
+            mask = mask[:, None, :, None] * context_mask[:, None, None, :]
             dots = dots.masked_fill(~mask, mask_value)
 
         # attention
@@ -437,6 +432,7 @@ class Alphafold2(nn.Module):
         num_embedds = constants.NUM_EMBEDDS_TR,
         max_num_msas = constants.MAX_NUM_MSA,
         max_num_templates = constants.MAX_NUM_TEMPLATES,
+        extra_msa_evoformer_layers = 4,
         attn_dropout = 0.,
         ff_dropout = 0.,
         sparse_self_attn = False,
@@ -466,6 +462,18 @@ class Alphafold2(nn.Module):
         self.max_rel_dist = max_rel_dist
         self.pos_emb = nn.Embedding(max_rel_dist * 2 + 1, dim)
 
+        # extra msa embedding
+
+        self.extra_msa_evoformer = Evoformer(
+            dim = dim,
+            depth = extra_msa_evoformer_layers,
+            seq_len = max_seq_len,
+            heads = heads,
+            dim_head = dim_head,
+            attn_dropout = attn_dropout,
+            ff_dropout = ff_dropout
+        )
+
         # template embedding
 
         self.to_template_embed = nn.Linear(template_dim, dim)
@@ -478,6 +486,13 @@ class Alphafold2(nn.Module):
             seq_len = max_seq_len
         )
 
+        self.template_pointwise_attn = Attention(
+            dim = dim,
+            dim_head = dim_head,
+            heads = heads,
+            dropout = attn_dropout
+        )
+
         # projection for angles, if needed
 
         self.predict_angles = predict_angles
@@ -599,13 +614,13 @@ class Alphafold2(nn.Module):
             m = m + rearrange(x, 'b n d -> b () n d')
 
             # get msa_mask to all ones if none was passed
-            msa_mask = default(msa_mask, torch.ones_like(msa).bool())
+            msa_mask = default(msa_mask, lambda: torch.ones_like(msa).bool())
 
         elif exists(embedds):
             m = self.embedd_project(embedds)
             
             # get msa_mask to all ones if none was passed
-            msa_mask = default(msa_mask, torch.ones_like(embedds[..., -1]).bool())
+            msa_mask = default(msa_mask, lambda: torch.ones_like(embedds[..., -1]).bool())
         else:
             raise Error('either MSA or embeds must be given')
 
@@ -617,7 +632,7 @@ class Alphafold2(nn.Module):
 
         # add relative positional embedding
 
-        seq_index = default(seq_index, torch.arange(n, device = device))
+        seq_index = default(seq_index, lambda: torch.arange(n, device = device))
         seq_rel_dist = rearrange(seq_index, 'i -> () i ()') - rearrange(seq_index, 'j -> () () j')
         seq_rel_dist = seq_rel_dist.clamp(-self.max_rel_dist, self.max_rel_dist) + self.max_rel_dist
         rel_pos_emb = self.pos_emb(seq_rel_dist)
@@ -630,21 +645,49 @@ class Alphafold2(nn.Module):
             # embed template
 
             t = self.to_template_embed(templates_feats)
-            templates_mask_crossed = rearrange(templates_mask, 'b t i -> b t i ()') * rearrange(templates_mask, 'b t j -> b t () j')
+            t_mask_crossed = rearrange(templates_mask, 'b t i -> b t i ()') * rearrange(templates_mask, 'b t j -> b t () j')
 
             t = rearrange(t, 'b t ... -> (b t) ...')
-            templates_mask_crossed = rearrange(templates_mask_crossed, 'b t ... -> (b t) ...')
+            t_mask_crossed = rearrange(t_mask_crossed, 'b t ... -> (b t) ...')
 
             for _ in range(self.template_embed_layers):
-                t = self.template_pairwise_embedder(t, mask = templates_mask_crossed) + t
+                t = self.template_pairwise_embedder(t, mask = t_mask_crossed) + t
 
             t = rearrange(t, '(b t) ... -> b t ...', t = num_templates)
-            templates_mask_crossed = rearrange(templates_mask_crossed, '(b t) ... -> b t ...', t = num_templates)
+            t_mask_crossed = rearrange(t_mask_crossed, '(b t) ... -> b t ...', t = num_templates)
 
             # template pos emb
 
-            assert t.shape[-2:] == x.shape[-2:]
-            x = x + t.mean(dim = 1)
+            x_point = rearrange(x, 'b i j d -> (b i j) () d')
+            t_point = rearrange(t, 'b t i j d -> (b i j) t d')
+            x_mask_point = rearrange(x_mask, 'b i j -> (b i j) ()')
+            t_mask_point = rearrange(t_mask_crossed, 'b t i j -> (b i j) t')
+
+            template_pooled = self.template_pointwise_attn(
+                x_point,
+                context = t_point,
+                mask = x_mask_point,
+                context_mask = t_mask_point
+            )
+
+            template_pooled_mask = rearrange(t_mask_point.sum(dim = -1) > 0, 'b -> b () ()')
+            template_pooled = template_pooled * template_pooled_mask
+
+            template_pooled = rearrange(template_pooled, '(b i j) () d -> b i j d', i = n, j = n)
+            x = x + template_pooled
+
+        # embed extra msa, if present
+
+        if exists(extra_msa):
+            extra_m = self.token_emb(msa)
+            extra_msa_mask = default(extra_msa_mask, torch.ones_like(extra_m).bool())
+
+            x, extra_m = self.extra_msa_evoformer(
+                x,
+                extra_m,
+                mask = x_mask,
+                msa_mask = extra_msa_mask
+            )
 
         # trunk
 
